import asyncio
import re
import sqlite3
from typing import Any, Callable

import asqlite

INBOX_STAFF_MENTION_PATTERN = re.compile(r"<@\d+>|<@&\d+>")


class DatabaseClient:
    """Provides an API for making common queries with an :class:`asqlite.Connection`."""

    def __init__(self, conn: asqlite.Connection) -> None:
        self.conn = conn

    # User methods

    async def add_user(self, user_id: int) -> None:
        """Add a user to the database.

        Existing users are ignored.

        """
        await self.conn.execute("INSERT OR IGNORE INTO user (id) VALUES (?)", user_id)

    # Guild methods

    async def add_guild(self, guild_id: int) -> None:
        """Add a guild to the database.

        Existing guilds are ignored.

        """
        await self.conn.execute("INSERT OR IGNORE INTO guild (id) VALUES (?)", guild_id)

    # Member methods

    async def add_member(self, user_id: int, guild_id: int) -> None:
        """Add a member to the database.

        Existing members and guilds are ignored.

        """
        await self.add_user(user_id)
        await self.add_guild(guild_id)
        await self.conn.execute(
            "INSERT OR IGNORE INTO member (guild_id, user_id) VALUES (?, ?)",
            guild_id,
            user_id,
        )

    async def add_user_or_member(
        self,
        user_id: int,
        *,
        guild_id: int | None = None,
    ) -> None:
        """Add a user/member to the database.

        Existing users, members, and guilds are ignored.

        """
        if guild_id is None:
            await self.add_user(user_id)
        else:
            await self.add_member(user_id, guild_id)

    # Channel methods

    async def add_channel(
        self,
        channel_id: int,
        *,
        guild_id: int | None = None,
    ) -> None:
        """Add a channel to the database.

        Existing channels and guilds are ignored.

        """
        if guild_id is not None:
            await self.add_guild(guild_id)

        await self.conn.execute(
            "INSERT OR IGNORE INTO channel (id, guild_id) VALUES (?, ?)",
            channel_id,
            guild_id,
        )

    # Message methods

    async def add_message(
        self,
        message_id: int,
        channel_id: int,
        *,
        guild_id: int | None = None,
    ) -> None:
        """Add a message to the database.

        Existing messages, channels, and guilds are ignored.

        """
        await self.add_channel(channel_id, guild_id=guild_id)
        await self.conn.execute(
            "INSERT OR IGNORE INTO message (id, channel_id) VALUES (?, ?)",
            message_id,
            channel_id,
        )

    # Inbox methods

    async def add_inbox(
        self,
        message_id: int,
        channel_id: int,
        *,
        guild_id: int | None = None,
    ) -> None:
        """Add an inbox to the database.

        :raises sqlite3.IntegrityError: The inbox already exists.

        """
        await self.add_message(message_id, channel_id, guild_id=guild_id)
        await self.conn.execute("INSERT INTO inbox (id) VALUES (?)", message_id)

    async def get_inbox_starter_content(self, inbox_id: int) -> str:
        """Get the starter content for an inbox.

        :returns: The starter content, if any.

        """
        row = await self.conn.fetchone(
            "SELECT starter_content FROM inbox WHERE id = ?",
            inbox_id,
        )
        assert row is not None
        return row[0]

    async def set_inbox_starter_content(
        self,
        inbox_id: int,
        starter_content: str,
    ) -> None:
        """Set the starter content for an inbox."""
        await self.conn.execute(
            "UPDATE inbox SET starter_content = ? WHERE id = ?",
            starter_content,
            inbox_id,
        )

    async def get_inbox_default_ticket_name(self, inbox_id: int) -> str:
        """Get the default ticket name for an inbox.

        :returns: The default ticket name, if any.

        """
        row = await self.conn.fetchone(
            "SELECT default_ticket_name FROM inbox WHERE id = ?",
            inbox_id,
        )
        assert row is not None
        return row[0]

    async def set_inbox_default_ticket_name(
        self,
        inbox_id: int,
        default_ticket_name: str,
    ) -> None:
        """Set the default ticket name for an inbox."""
        await self.conn.execute(
            "UPDATE inbox SET default_ticket_name = ? WHERE id = ?",
            default_ticket_name,
            inbox_id,
        )

    async def increment_inbox_counter(self, inbox_id: int) -> int:
        """Increment the counter for an inbox.

        :returns: The new value of the counter.

        """
        row = await self.conn.fetchone(
            "UPDATE inbox SET counter = counter + 1 WHERE id = ? RETURNING counter",
            inbox_id,
        )
        assert row is not None
        return row[0]

    async def add_inbox_staff(self, inbox_id: int, mention: str) -> None:
        """Add an inbox staff to the database.

        :raises ValueError:
            The mention string does not match a user or role format.
        :raises sqlite3.IntegrityError:
            Either the inbox does not exist or the mention already exists.

        """
        if not INBOX_STAFF_MENTION_PATTERN.fullmatch(mention):
            raise ValueError(f"Invalid user/role mention: {mention!r}")

        await self.conn.execute(
            "INSERT INTO inbox_staff (inbox_id, mention) VALUES (?, ?)",
            inbox_id,
            mention,
        )

    async def get_inbox_staff(self, inbox_id: int) -> list[str]:
        """Get all staff mentions for an inbox.

        If the inbox does not exist, an empty list is returned.

        :returns: A list of staff mentions.

        """
        rows = await self.conn.fetchall(
            "SELECT mention FROM inbox_staff WHERE inbox_id = ?",
            inbox_id,
        )
        return [row[0] for row in rows]

    async def remove_inbox_staff(self, inbox_id: int, mention: str) -> bool:
        """Remove an inbox staff from the database.

        :returns: True if the given inbox staff existed, False otherwise.

        """
        row = await self.conn.fetchone(
            "DELETE FROM inbox_staff WHERE inbox_id = ? AND mention = ? RETURNING 1",
            inbox_id,
            mention,
        )
        return row is not None

    # Ticket methods

    async def add_ticket(
        self,
        *,
        ticket_id: int,
        inbox_id: int,
        owner_id: int,
        guild_id: int,
    ) -> None:
        """Add a ticket to the database.

        :raises sqlite3.IntegrityError: The ticket already exists.

        """
        await self.add_member(owner_id, guild_id)
        await self.add_channel(ticket_id, guild_id=guild_id)
        await self.add_channel(inbox_id, guild_id=guild_id)
        await self.conn.execute(
            "INSERT INTO ticket (id, inbox_id, owner_id) VALUES (?, ?, ?)",
            ticket_id,
            inbox_id,
            owner_id,
        )

    async def count_matching_tickets(self, ticket_ids: list[int]) -> int:
        """Count the number of ticket IDs that exist in the database."""
        placeholders = ", ".join("?" * len(ticket_ids))
        row = await self.conn.fetchone(
            f"SELECT COUNT(*) FROM ticket WHERE id IN ({placeholders})",
            *ticket_ids,
        )
        assert row is not None
        return row[0]


def connect(
    database: str | bytes,
    *,
    init: Callable[[sqlite3.Connection], None] | None = None,
    timeout: float | None = None,
    loop: asyncio.AbstractEventLoop | None = None,
    **kwargs: Any,
) -> asqlite._ContextManagerMixin[sqlite3.Connection, asqlite.Connection]:
    # This is a monkeypatch of asqlite's connect() function since the init=
    # callback doesn't run before asqlite's own pragmas, which caused
    # connections to fail on encrypted databases.
    loop = loop or asyncio.get_event_loop()
    queue = asqlite._Worker(loop=loop)
    queue.start()

    def factory(con: sqlite3.Connection) -> asqlite.Connection:
        return asqlite.Connection(con, queue)

    if init is not None:

        def new_connect(db: str | bytes, **kwargs: Any) -> sqlite3.Connection:
            # init order flipped, pragmas copied directly from _connect_pragmas()
            conn = sqlite3.connect(db, **kwargs)
            init(conn)
            conn.execute("pragma journal_mode=wal")
            conn.execute("pragma foreign_keys=ON")
            conn.isolation_level = None
            conn.row_factory = sqlite3.Row
            return conn

    else:
        new_connect = asqlite._connect_pragmas  # type: ignore

    return asqlite._ContextManagerMixin(
        queue,
        factory,
        new_connect,
        database,
        timeout=timeout,
        **kwargs,
    )



from __future__ import annotations

import importlib.resources
import tomllib
from pathlib import Path
from typing import IO, TYPE_CHECKING, Annotated, Any, Literal, Protocol

from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    SecretStr,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapValidator,
)

from .appdirs import expand_app_dirs

if TYPE_CHECKING:
    import discord

assert __package__ is not None
_package_files = importlib.resources.files(__package__)
CONFIG_DEFAULT_RESOURCE = _package_files.joinpath("config_default.toml")


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


# https://docs.pydantic.dev/usage/settings/
class Settings(_BaseModel):
    bot: SettingsBot
    db: SettingsDB


class SettingsBot(_BaseModel):
    allow_jishaku: bool
    extensions: list[str]
    inbox: SettingsBotInbox
    intents: SettingsBotIntents
    token: str


class SettingsBotInbox(_BaseModel):
    max_attachment_size: int
    """The max cumulative size allowed for an inbox message's attachments."""


class SettingsBotIntents(_BaseModel):
    """The intents used when connecting to the Discord gateway.

    .. seealso:: https://discordpy.readthedocs.io/en/stable/api.html#intents

    """

    model_config = ConfigDict(extra="allow")

    def create_intents(self) -> discord.Intents:
        import discord

        return discord.Intents(**self.model_dump())


def expand_app_dirs_strict(s: str) -> str:
    try:
        return expand_app_dirs(s, strict=True)
    except KeyError as e:
        raise ValueError(f"{e.args[0]} not a valid variable") from None


def check_pragma_statement(s: SecretStr) -> SecretStr:
    if not s.get_secret_value().strip().lower().startswith("pragma"):
        raise ValueError("statement must start with PRAGMA")
    return s


def check_single_statement(s: SecretStr) -> SecretStr:
    s_new = s.get_secret_value().rstrip(";")
    if ";" in s_new:
        raise ValueError("string must consist of a single statement")
    return SecretStr(s_new)


def pass_through_empty_string(
    s: str,
    handler: ValidatorFunctionWrapHandler,
    info: ValidationInfo,
) -> SecretStr:
    if s == "":
        return SecretStr(s)
    return handler(s)


class SettingsDB(_BaseModel):
    path: Annotated[Path, BeforeValidator(expand_app_dirs_strict)]
    """The filepath to write the bot's SQLite database to."""
    pragmas: list[
        Annotated[
            SecretStr,
            AfterValidator(check_single_statement),
            AfterValidator(check_pragma_statement),
        ]
    ]
    """A list of pragmas used to initialize the database."""
    key_template: Annotated[
        SecretStr,
        AfterValidator(check_single_statement),
        AfterValidator(check_pragma_statement),
        WrapValidator(pass_through_empty_string),
    ]
    """The pragma template used to prompt for the passphrase upon startup."""


Settings.model_rebuild()
SettingsBot.model_rebuild()


class OpenableBinary(Protocol):
    def open(self, __mode: Literal["rb"], /) -> IO[bytes]: ...


def _recursive_update(dest: dict, src: dict) -> None:
    for k, vsrc in src.items():
        vdest = dest.get(k)
        if isinstance(vdest, dict) and isinstance(vsrc, dict):
            _recursive_update(vdest, vsrc)
        else:
            dest[k] = vsrc


def _load_raw_config(path: OpenableBinary) -> dict[str, Any]:
    with path.open("rb") as f:
        return tomllib.load(f)


def load_default_config() -> Settings:
    """Loads the default configuration file.

    :returns: The settings that were parsed.
    :raises FileNotFoundError:
        The default configuration file could not be found.

    """
    data = _load_raw_config(CONFIG_DEFAULT_RESOURCE)
    return Settings.model_validate(data)


def load_config(path: Path, *, merge_default: bool = True) -> Settings:
    """Loads the bot configuration file.

    :param merge_default:
        If True, the default configuration file will be used as a base
        and the normal configuration is applied on top of it,
        if it exists.
    :returns: The settings that were parsed.
    :raises FileNotFoundError:
        Either the configuration file could not be found,
        or the default configuration file could not be found.

    """
    if not merge_default:
        data = _load_raw_config(path)
    else:
        data = _load_raw_config(CONFIG_DEFAULT_RESOURCE)
        overwrites = _load_raw_config(path)
        _recursive_update(data, overwrites)

    return Settings.model_validate(data)



import argparse
import asyncio
import functools
import getpass
import importlib.metadata
import logging
import os
import sys
from pathlib import Path
from typing import Type

from pydantic import SecretStr

from . import __version__
from .appdirs import APP_DIRS
from .bot import Bot, StartupFlags
from .config import load_config
from .logging import configure_logging

log = logging.getLogger(__package__)

USER_CONFIG = APP_DIRS.user_config_path / "config.toml"


def suppress(*exceptions: Type[BaseException]):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exceptions:
                return

        return wrapper

    return decorator


@suppress(KeyboardInterrupt)
def main() -> None:
    parser = argparse.ArgumentParser(
        prog=__package__,
        description=importlib.metadata.metadata("theticketbot")["Summary"],
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase logging verbosity",
    )
    parser.add_argument(
        "--config-file",
        help="The config file to load",
        type=Path,
    )
    commands = parser.add_mutually_exclusive_group()
    commands.add_argument(
        "--sync",
        action="store_const",
        const=StartupFlags.SYNC | StartupFlags.CLOSE,
        default=StartupFlags(0),
        dest="startup_flags",
        help="Sync application commands at startup",
    )
    commands.add_argument(
        "--dump-config",
        action="store_true",
        help="Dump config file at startup",
    )

    args = parser.parse_args()
    config_file: Path | None = args.config_file
    startup_flags: StartupFlags = args.startup_flags

    # Configure logging early to capture our own initialization
    configure_logging(args.verbose)

    if args.dump_config:
        dump_config_and_exit(config_file)

    if config_file is None:
        config_file = find_config_file()

    temp_config_file: Path | None = None
    if config_file is None:
        config_file = temp_config_file = prompt_and_create_config_file()
        startup_flags = StartupFlags.SYNC

    bot = Bot(
        functools.partial(load_config, config_file),
        startup_flags=startup_flags,
    )

    check_outdated_database_path(bot.config.db.path, args.config_file)

    if bot.config.bot.token == "":
        sys.exit(
            "No bot token has been supplied by the config file.\n"
            "Please get a Bot Token from https://discord.com/developers/applications "
            "and add it to your configuration."
        )

    key_template = bot.config.db.key_template.get_secret_value()
    if key_template != "":
        key = getpass.getpass("Database Key: ")
        pragma = key_template.format(key)
        bot.key_pragma = SecretStr(pragma)

    log.info(f"Package version: {__version__}")
    asyncio.run(start(bot, temp_config_file))


def dump_config_and_exit(config_file: Path | None) -> None:
    if config_file is None:
        config_file = find_config_file()
    if config_file is None:
        sys.exit(
            "The config.toml file is not not present.\n"
            "Please run the bot normally to generate a configuration file,\n"
            "or write your own configuration file."
        )

    config = load_config(config_file)
    print(config_file)
    print(config.model_dump_json(indent=4, exclude={"bot": {"token"}}))
    sys.exit(1)


def find_config_file() -> Path | None:
    cwd_config = Path("config.toml")
    if USER_CONFIG.exists():
        if cwd_config.exists():
            log.warning(
                "\n"
                "Both the user-specific config.toml and CWD config.toml are present.\n"
                "Only the user config will be loaded.\n"
                "If you want to load the config.toml from CWD instead,\n"
                "add `--config-file config.toml` to your command-line arguments."
            )
        return USER_CONFIG
    elif cwd_config.exists():
        return cwd_config


def prompt_and_create_config_file() -> Path:
    print(
        f"No config.toml file was found in the current working directory.\n"
        f"A minimal config file will be written to:\n"
        f"\n"
        f"    {USER_CONFIG}\n"
        f"\n"
        f"Afterwards, the bot will automatically synchronize its application commands\n"
        f"and start itself. If an error occurs during this process, the config.toml\n"
        f"will be deleted so you can redo this.\n"
    )

    token = input_token()

    USER_CONFIG.parent.mkdir(parents=True, exist_ok=True)
    with USER_CONFIG.open("w") as f:
        f.write("[bot]\n")
        f.write(f'token = "{token}"\n')

    return USER_CONFIG


def input_token() -> str:
    import getpass
    import re

    while True:
        token = getpass.getpass("Bot Token: ").strip()
        if re.fullmatch(r"\w+\.\w+\.\S+", token) is not None:
            return token

        print(
            "This token does not appear to be valid. It should look something like:\n"
            "    MTI0NjgyNjg0MTIzMTMyNzI3NQ.GTIAZm.x2fbSNuYJgpAocvMM53ROlMC23NixWt-0NOjMc\n"
            "Please try again."
        )


def check_outdated_database_path(path: Path, config_file: str) -> None:
    if os.getenv("IGNORE_OUTDATED_DATABASE_PATHS") == "1":
        return

    outdated_paths = [Path("data/theticketbot.db")]
    for outdated in outdated_paths:
        if path.resolve() == outdated.resolve():
            return
        elif not outdated.exists():
            continue
        elif outdated.is_symlink():
            continue

        log.warning(
            f"\n"
            f"A database file was found at {outdated} but db.path is set to\n"
            f"{path}.\n"
            f"\n"
            f"It is recommended you move your database file to db.path\n"
            f"so theticketbot can continue using the database.\n"
            f"\n"
            f"If you want to keep using your database file at {outdated},\n"
            f'please set path = "{outdated}" in the [db] table\n'
            f"of your config file, {config_file}.\n"
            f"\n"
            f"If you are absolutely sure your database files are correctly placed,\n"
            f"you may suppress this warning by setting the environment variable\n"
            f"IGNORE_OUTDATED_DATABASE_PATHS=1."
        )
        return


async def start(bot: Bot, temp_config_file: Path | None) -> None:
    try:
        async with bot:
            await bot.login(bot.config.bot.token)

            # Token is valid, we no longer need to delete the temporary
            # config file if an error occurs now.
            temp_config_file = None

            await bot.connect(reconnect=True)
    except BaseException:
        if temp_config_file is not None:
            temp_config_file.unlink()
        raise


if __name__ == "__main__":
    main()



import datetime
import logging
import discord
from discord.ext import commands, tasks

from theticketbot.bot import Bot

log = logging.getLogger(__name__)


class Cleanup(commands.Cog):
    def __init__(self, bot: Bot):
        self.bot = bot
        self.cleanup_loop.start()

    # @commands.Cog.listener("on_guild_remove")
    # async def remove_guild(self, guild: discord.Guild):
    #     async with self.bot.acquire() as conn:
    #         await conn.execute("DELETE FROM guild WHERE id = ?", guild.id)
    #
    # In case the bot is unintentionally kicked, retain all tickets
    # until the next cleanup cycle

    @commands.Cog.listener("on_guild_channel_delete")
    async def remove_guild_channel(self, channel: discord.abc.GuildChannel):
        async with self.bot.acquire() as conn:
            await conn.execute("DELETE FROM channel WHERE id = ?", channel.id)

    @commands.Cog.listener("on_raw_thread_delete")
    async def remove_thread(self, payload: discord.RawThreadDeleteEvent):
        async with self.bot.acquire() as conn:
            await conn.execute("DELETE FROM channel WHERE id = ?", payload.thread_id)

    @commands.Cog.listener("on_raw_message_delete")
    async def remove_message(self, payload: discord.RawMessageDeleteEvent):
        async with self.bot.acquire() as conn:
            await conn.execute("DELETE FROM message WHERE id = ?", payload.message_id)

    @commands.Cog.listener("on_raw_bulk_message_delete")
    async def bulk_remove_messages(self, payload: discord.RawBulkMessageDeleteEvent):
        async with self.bot.acquire() as conn:
            await conn.executemany(
                "DELETE FROM message WHERE id = ?",
                [(id,) for id in payload.message_ids],
            )

    @tasks.loop(time=datetime.time(0, 0, tzinfo=datetime.timezone.utc))
    async def cleanup_loop(self) -> None:
        now = datetime.datetime.now(datetime.timezone.utc)
        if now.weekday() != 5:
            return

        await self.cleanup_guilds()

    async def cleanup_guilds(self) -> None:
        guild_ids = {guild.id for guild in self.bot.guilds}
        async with self.bot.acquire() as conn:
            rows = await conn.fetchall("SELECT id FROM guild")
            rows = {row[0] for row in rows}
            rows = rows - guild_ids
            rows = [(guild_id,) for guild_id in rows]
            await conn.executemany("DELETE FROM guild WHERE id = ?", rows)

        if len(rows) > 0:
            log.info("%d guilds cleaned up", rows)

    # NOTE: users are not removed by any event
    # NOTE: members are not removed by any event, members intent required
    # NOTE: rows can still accumulate during bot downtime


async def setup(bot: Bot):
    await bot.add_cog(Cleanup(bot))



import logging
import random
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, ClassVar, Generic, Type, TypeVar, cast

import discord
from discord import app_commands
from discord.app_commands import locale_str as _
from discord.ext import commands

from theticketbot.bot import Bot, Context
from theticketbot.errors import AppCommandResponse
from theticketbot.translator import translate

log = logging.getLogger(__name__)

T = TypeVar("T")


async def append_error_code(ctx: object, content: str, error_code: str) -> str:
    trailer = _(
        # Message appended to some error responses caused by issues in the bot
        # {0}: the error code to be reported
        # {1}: the maintainer's mention
        "Error code: {0}\n"
        "If assistance is needed, please contact {1}."
    )
    trailer = await maybe_translate(ctx, trailer)
    trailer = trailer.format(error_code, get_owner_mention(ctx))
    return f"{content}\n{trailer}"


def get_owner_mention(ctx: object) -> str:
    if isinstance(ctx, discord.Interaction):
        ctx = cast(discord.Interaction[Bot], ctx)
        bot = ctx.client
    elif isinstance(ctx, commands.Context):
        ctx = cast(commands.Context[Bot], ctx)
        bot = ctx.bot
    else:
        # FIXME: should ctx really be generic?
        raise TypeError(f"Unsupported context type {type(ctx).__name__}")

    if bot.application is None:
        raise RuntimeError("Owner mention not available, bot must be logged in")

    return bot.application.owner.mention


def generate_error_code():
    return "".join(random.choices("0123456789ABCDEF", k=4))


async def maybe_translate(ctx: object, message: str | app_commands.locale_str) -> str:
    if isinstance(message, str):
        return message
    elif isinstance(ctx, discord.Interaction):
        return await translate(message, ctx)
    elif isinstance(ctx, commands.Context) and ctx.guild is not None:
        locale = ctx.guild.preferred_locale
        return await translate(message, ctx.bot, locale=locale)
    else:
        return message.message


@dataclass
class ErrorResponse:
    exc_types: Type[Exception] | tuple[Type[Exception], ...]
    content: str | app_commands.locale_str | None
    show_traceback: bool

    async def format(self, ctx: Any, error: Exception) -> str | None:
        if self.content is not None:
            content = await maybe_translate(ctx, self.content)
            return content.format(error)


class AppCommandErrorResponse(ErrorResponse):
    """A dedicated error formatter for :exc:`AppCommandResponse`."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            AppCommandResponse,
            None,
            *args,
            **kwargs,
        )

    async def format(self, ctx: discord.Interaction, error: Exception) -> str | None:
        assert isinstance(error, AppCommandResponse)
        message = error.message
        if isinstance(message, app_commands.locale_str):
            message = await translate(message, ctx, error.data)
        return message


class ErrorHandler(ABC, Generic[T]):
    responses: ClassVar[list[ErrorResponse]]

    @abstractmethod
    def get_command_name(self, ctx: T, /) -> str | None:
        """Returns the name of the command that was invoked, if any."""

    @abstractmethod
    def get_original_error(self, error: Exception, /) -> BaseException | None:
        """Returns the original error of an exception if available.

        Whatever exception is returned here will be used for logging
        the traceback. If None is returned, ``error`` is logged instead.

        """

    @abstractmethod
    async def send(self, ctx: T, content: str, /) -> None:
        """Sends a response to the command invoker."""

    @abstractmethod
    def should_handle(self, ctx: T, error: Exception, /) -> bool:
        """Checks if the given error should be handled."""

    async def handle(self, ctx: T, error: Exception, /) -> None:
        if not self.should_handle(ctx, error):
            return

        error_code = generate_error_code()

        for resp in self.responses:
            if isinstance(error, resp.exc_types):
                content = await resp.format(ctx, error)
                break
        else:
            raise ValueError(f"Unable to handle exception: {error!r}")

        if resp.show_traceback:
            self._log_error(ctx, error, error_code)

        if content is not None:
            if resp.show_traceback:
                content = await append_error_code(ctx, content, error_code)
            await self.send(ctx, content)

    def _log_error(self, ctx: T, error: Exception, error_code: str) -> None:
        error_name = type(error).__name__
        command = self.get_command_name(ctx) or "no command"
        original = self.get_original_error(error) or error

        log.error(
            f'{error_name} while invoking "{command}" (code: {error_code})',
            exc_info=original,
        )


class PrefixErrorHandler(ErrorHandler[Context]):
    responses = [
        ErrorResponse(
            (commands.CommandNotFound, commands.DisabledCommand),
            content=None,
            show_traceback=False,
        ),
        ErrorResponse(
            commands.CommandOnCooldown,
            # Error response for command on cooldown
            # {0}: an exception containing the duration to wait in .retry_after
            _("This command is on cooldown for {0.retry_after:.1f}s."),
            show_traceback=False,
        ),
        ErrorResponse(
            commands.MaxConcurrencyReached,
            # Error response for exceeding maximum concurrent users of a command
            _("Too many people are using this command. Please try again later."),
            show_traceback=False,
        ),
        ErrorResponse(
            commands.CheckFailure,
            # Error response for not passing all checks required to use a command
            _("One or more checks failed for this command."),
            show_traceback=False,
        ),
        ErrorResponse(
            commands.UserInputError,
            # Error response for failing to parse the user's input
            # {0}: the error description
            _("An error occurred with your input: ```py\n{0}```"),
            show_traceback=True,
        ),
        ErrorResponse(
            Exception,
            # Error response for an unexpected failure in a command
            _("An unknown error occurred while running this command."),
            show_traceback=True,
        ),
    ]

    def get_command_name(self, ctx: Context) -> str | None:
        if ctx.command is not None:
            return ctx.command.qualified_name

    def get_original_error(self, error: Exception) -> BaseException | None:
        if isinstance(error, (commands.CommandInvokeError, commands.ConversionError)):
            return error.original
        elif isinstance(error, commands.HybridCommandError):
            return TreeErrorHandler().get_original_error(error)

    async def send(self, ctx: Context, content: str) -> None:
        await ctx.send(
            content,
            allowed_mentions=discord.AllowedMentions.none(),
            ephemeral=True,
        )

    def should_handle(self, ctx: Context, error: Exception, /) -> bool:
        return not getattr(ctx, "handled", False)


class TreeErrorHandler(ErrorHandler[discord.Interaction]):
    responses = [
        AppCommandErrorResponse(show_traceback=False),
        ErrorResponse(
            app_commands.CommandOnCooldown,
            # Error response for command on cooldown
            # {0}: an exception containing the duration to wait in .retry_after
            _("This command is on cooldown for {0.retry_after:.1f}s."),
            show_traceback=False,
        ),
        ErrorResponse(
            app_commands.CheckFailure,
            # Error response for not passing all checks required to use a command
            _("One or more checks failed for this command."),
            show_traceback=False,
        ),
        ErrorResponse(
            app_commands.TransformerError,
            # Error response for failing to parse the user's input
            # {0}: the error description
            _("An error occurred with your input: ```py\n{0}```"),
            show_traceback=True,
        ),
        ErrorResponse(
            app_commands.CommandNotFound,
            # Error response for using a slash command not recognized by the bot
            _("The bot currently does not recognize this command."),
            show_traceback=True,
        ),
        ErrorResponse(
            Exception,
            # Error response for an unexpected failure in a command
            _("An unknown error occurred while running this command."),
            show_traceback=True,
        ),
    ]

    def get_command_name(self, interaction: discord.Interaction) -> str | None:
        if interaction.command is not None:
            return "/" + interaction.command.qualified_name

    def get_original_error(self, error: Exception) -> BaseException | None:
        if isinstance(error, app_commands.CommandInvokeError):
            return error.original
        elif isinstance(error, app_commands.TransformerError):
            return error.__cause__

    async def send(self, interaction: discord.Interaction, content: str) -> None:
        kwargs = {
            "content": content,
            "allowed_mentions": discord.AllowedMentions.none(),
            "ephemeral": True,
        }
        if interaction.response.is_done():
            await interaction.followup.send(**kwargs)
        else:
            await interaction.response.send_message(**kwargs)

    def should_handle(
        self,
        interaction: discord.Interaction,
        error: Exception,
        /,
    ) -> bool:
        return not interaction.extras.get("handled", False)


class Errors(commands.Cog):
    def __init__(self, bot: Bot):
        self.bot = bot
        self.prefix_handler = PrefixErrorHandler()
        self.tree_handler = TreeErrorHandler()
        self.setup_events()

    async def cog_unload(self):
        self.teardown_events()

    def setup_events(self):
        self._old_command_error = self.bot.on_command_error
        self.bot.on_command_error = self.prefix_handler.handle  # type: ignore

        self._old_tree_error = self.bot.tree.on_error
        self.bot.tree.error(self.tree_handler.handle)

    def teardown_events(self):
        self.bot.on_command_error = self._old_command_error
        self.bot.tree.error(self._old_tree_error)  # type: ignore


async def setup(bot: Bot):
    await bot.add_cog(Errors(bot))



"""Re-invokes xgettext on source packages and merges them into .po/.pot files."""

import argparse
import subprocess
from contextlib import contextmanager, nullcontext
from pathlib import Path, PurePosixPath
from typing import Generator, Iterable


@contextmanager
def temporary_po_from_source(
    source_files: Iterable[Path],
    output_path: Path,
) -> Generator[Path, None, None]:
    subprocess.check_call(
        [
            "xgettext",
            # Extract comments from source
            # https://www.gnu.org/software/gettext/manual/gettext.html#index-_002dc_002c-xgettext-option
            "--add-comments",
            "-o",
            # Normalize generated source file references in POSIX style
            PurePosixPath(output_path),
            *(PurePosixPath(p) for p in source_files),
        ],
    )

    try:
        # Hide CHARSET warning by defaulting to utf-8
        content_type_temp = rb'"Content-Type: text/plain; charset=CHARSET\n"'
        content_type_utf8 = rb'"Content-Type: text/plain; charset=UTF-8\n"'
        content_pot = output_path.read_bytes()
        content_pot = content_pot.replace(content_type_temp, content_type_utf8)
        output_path.write_bytes(content_pot)

        yield output_path
    finally:
        output_path.unlink(missing_ok=True)


parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "-t",
    "--template",
    action="store_true",
    dest="merge_template",
    help="Instead of generating from source, use the POT file to merge",
)
args = parser.parse_args()

for package_path in Path("src").iterdir():
    if not package_path.is_dir():
        continue

    pot_files = tuple(package_path.glob("*.pot"))
    if not pot_files:
        continue

    print(package_path)

    source_files: list[Path] = []
    source_files.extend(package_path.rglob("*.py"))
    if not source_files:
        continue

    po_paths: list[Path] = []
    po_paths.extend(package_path.rglob("*.po"))
    if not args.merge_template:
        po_paths.extend(package_path.rglob("*.pot"))
    if not po_paths:
        continue

    if args.merge_template:
        merging_po_cm = nullcontext(pot_files[0])
        print(f"Merging from {pot_files[0]}...")
    else:
        merging_po_cm = temporary_po_from_source(
            source_files,
            output_path=package_path / "messages.po.merging",
        )
        print("Generating PO from source to merge...")

    with merging_po_cm as merging_po:
        for po_path in po_paths:
            print(po_path)
            subprocess.check_call(["msgmerge", po_path, merging_po, "-o", po_path])



import time
from dataclasses import dataclass
from typing import Any, Awaitable, Callable

import discord
from discord import app_commands
from discord.app_commands import locale_str as _
from discord.ext import commands, tasks

from theticketbot.bot import Bot
from theticketbot.translator import translate

MessageCallback = Callable[[discord.Interaction, discord.Message], Awaitable[Any]]


@dataclass
class MessageCommand:
    timestamp: float
    callback: MessageCallback


class Select(commands.Cog):
    CLEANUP_INTERVAL = 60
    CLEANUP_EXPIRED_AFTER = 300
    MESSAGE_EXPIRES_AFTER = 180

    def __init__(self, bot: Bot):
        self.bot = bot
        self.cog_menus = (
            app_commands.ContextMenu(
                # Command name
                name=_("Select this message"),
                callback=self.on_message_selected,
            ),
        )

        for menu in self.cog_menus:
            bot.tree.add_command(menu)

        self._message_commands: dict[tuple[int, int], MessageCommand] = {}
        self.cleanup_loop.start()

    def set_message_callback(
        self,
        guild_id: int,
        user_id: int,
        callback: MessageCallback,
    ) -> None:
        """Set the next message callback for the given user."""
        key = (guild_id, user_id)
        self._message_commands[key] = MessageCommand(time.monotonic(), callback)

    async def cog_unload(self) -> None:
        for menu in self.cog_menus:
            self.bot.tree.remove_command(menu.name, type=menu.type)

    @app_commands.default_permissions(manage_guild=True)
    @app_commands.guild_only()
    async def on_message_selected(
        self,
        interaction: discord.Interaction,
        message: discord.Message,
    ):
        assert interaction.guild is not None
        key = (interaction.guild.id, interaction.user.id)
        command = self._message_commands.get(key)

        if command is None:
            content = _(
                # Message sent when selecting a message without a command
                "You can't select a message right now! "
                "Please use a command that asks for a message first."
            )
            content = await translate(content, interaction)
            return await interaction.response.send_message(content, ephemeral=True)
        elif time.monotonic() > command.timestamp + self.MESSAGE_EXPIRES_AFTER:
            content = _(
                # Message sent when selecting a message too long after their last command
                "Sorry, your last command has expired. "
                "Please use a command again and then select this message."
            )
            content = await translate(content, interaction)
            return await interaction.response.send_message(content, ephemeral=True)

        del self._message_commands[key]
        try:
            await command.callback(interaction, message)
        except BaseException:
            command.timestamp = time.monotonic()
            self._message_commands.setdefault(key, command)
            raise

    @tasks.loop(seconds=CLEANUP_INTERVAL)
    async def cleanup_loop(self) -> None:
        now = time.monotonic()

        to_remove: list[tuple[int, int]] = []
        cleanup_after = self.MESSAGE_EXPIRES_AFTER + self.CLEANUP_EXPIRED_AFTER
        for key, queue in self._message_commands.items():
            if now > queue.timestamp + cleanup_after:
                to_remove.append(key)

        for key in to_remove:
            del self._message_commands[key]


async def setup(bot: Bot):
    await bot.add_cog(Select(bot))



from typing import Any

from discord import app_commands


class AppCommandResponse(app_commands.AppCommandError):
    """An exception used for sending a message directly to the user.

    Parameters
    ----------
    message: str | app_commands.locale_str
        The message to be shown to the user. Can be a localized string.
    data: Any
        If message is a localized string, this data will be passed to
        the bot's translator.

    """

    def __init__(
        self,
        message: str | app_commands.locale_str,
        data: Any = None,
    ) -> None:
        super().__init__(str(message))
        self.message = message
        self.data = data



from __future__ import annotations

import gettext
import importlib.resources
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator

import discord
from discord import app_commands
from discord.app_commands import TranslationContextLocation
from discord.ext import commands

if TYPE_CHECKING:
    from .bot import Bot

assert __package__ is not None
_LOCALES_PATH = Path(str(importlib.resources.files(__package__).joinpath("locales")))
DOMAIN = "theticketbot"

log = logging.getLogger(__name__)


def locale_to_gnu(locale: discord.Locale) -> str:
    return str(locale).replace("-", "_")


def yield_mo_paths() -> Iterator[Path]:
    if not _LOCALES_PATH.is_dir():
        return

    for locale in _LOCALES_PATH.iterdir():
        lc_messages = locale / "LC_MESSAGES"
        if not lc_messages.is_dir():
            continue

        yield from lc_messages.glob("*.mo")


class EmptyTranslations(gettext.NullTranslations):
    """Returns an empty message to indicate no translation is available."""

    def gettext(self, message: str) -> str:
        return ""

    def ngettext(self, msgid1: str, msgid2: str, n: int) -> str:
        return ""

    def pgettext(self, context: str, message: str) -> str:
        return ""

    def npgettext(self, context: str, msgid1: str, msgid2: str, n: int) -> str:
        return ""


class GettextTranslator(app_commands.Translator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if not any(yield_mo_paths()):
            log.warning("No compiled localizations detected")

    async def translate(
        self,
        string: app_commands.locale_str,
        locale: discord.Locale,
        context: app_commands.TranslationContextTypes,
    ) -> str | None:
        try:
            t = gettext.translation(
                domain=DOMAIN,
                localedir=str(_LOCALES_PATH),
                languages=(locale_to_gnu(locale), "en_US"),
            )
        except OSError:
            return

        t.add_fallback(EmptyTranslations())

        plural: str | None = string.extras.get("plural")
        if plural is not None:
            if context.location == TranslationContextLocation.choice_name:
                quantity = context.data.value
            else:
                quantity = context.data
            assert isinstance(quantity, int)
            translated = t.ngettext(string.message, plural, quantity)
        else:
            translated = t.gettext(string.message)

        if context.location == TranslationContextLocation.choice_name:
            translated = translated.format(context.data.value)

        return translated or None


def plural_locale_str(
    singular: str,
    plural: str,
    **kwargs,
) -> app_commands.locale_str:
    """A shorthand for defining a string with singular and plural variants.

    This should be aliased to ngettext, ungettext, or dngettext
    so the `xgettext` program can recognize the function.

    """
    return app_commands.locale_str(singular, plural=plural, **kwargs)


async def translate(
    message: app_commands.locale_str,
    obj: Bot | discord.Interaction,
    locale: discord.Locale | None = None,
    data: Any = None,
) -> str:
    """A shorthand for translating a message.

    Unlike the methods built into discord.py, this will use the original message
    if a translation could not be found.

    """
    if isinstance(obj, commands.Bot):
        if locale is None:
            return str(message)

        assert obj.tree.translator is not None
        context = app_commands.TranslationContext(
            location=app_commands.TranslationContextLocation.other,
            data=data,
        )
        translated = await obj.tree.translator.translate(
            message,
            locale=locale,
            context=context,
        )
    else:
        translated = await obj.translate(
            message,
            data=data,
            locale=locale or obj.locale,
        )

    return translated or str(message)



"""Iterates through .po files in source packages and invokes msgfmt on them."""

import subprocess
from pathlib import Path

for po_path in Path("src").rglob("*.po"):
    print(po_path)
    output = po_path.with_suffix(".mo")
    subprocess.check_call(["msgfmt", "-o", output, po_path])



from __future__ import annotations

import contextlib
import importlib.metadata
import logging
import sqlite3
import sys
from enum import Flag, auto
from typing import TYPE_CHECKING, AsyncGenerator, Callable, cast

import asqlite
import discord
from discord.ext import commands
from pydantic import SecretStr

from . import database
from .migrations import run_default_migrations
from .translator import GettextTranslator

if TYPE_CHECKING:
    from .cogs.select import MessageCallback, Select
    from .config import Settings

log = logging.getLogger(__name__)


class StartupFlags(Flag):
    SYNC = auto()
    CLOSE = auto()


# https://discordpy.readthedocs.io/en/stable/ext/commands/api.html
class Bot(commands.Bot):
    key_pragma: SecretStr | None
    """The key pragma to append onto the database pragmas.

    This should be assigned after construction.

    """

    def __init__(
        self,
        config_refresher: Callable[[], Settings],
        *,
        startup_flags: StartupFlags,
    ):
        self._config_refresher = config_refresher
        config = self.refresh_config()

        self.startup_flags = startup_flags
        self.key_pragma = None

        super().__init__(
            chunk_guilds_at_startup=False,
            command_prefix=commands.when_mentioned,
            intents=config.bot.intents.create_intents(),
            member_cache_flags=discord.MemberCacheFlags.none(),
            strip_after_prefix=True,
        )

    @contextlib.asynccontextmanager
    async def acquire(
        self,
        *,
        transaction: bool = True,
    ) -> AsyncGenerator[asqlite.Connection, None]:
        """Acquire a connection to the database.

        :param transaction: If True, a transaction is opened as well.

        """
        path = str(self.config.db.path)
        init = lambda conn: self._run_config_pragmas(conn)
        async with database.connect(path, init=init) as conn:
            if not transaction:
                yield conn
            else:
                async with conn.transaction():
                    yield conn

    def _run_config_pragmas(self, conn: sqlite3.Connection) -> None:
        pragmas = [p.get_secret_value() for p in self.config.db.pragmas]
        if self.key_pragma is not None:
            pragmas.append(self.key_pragma.get_secret_value())
        conn.executescript(";\n".join(pragmas))

    def set_message_callback(
        self,
        guild_id: int,
        user_id: int,
        callback: MessageCallback,
    ) -> None:
        """Set the next message callback for the given user."""
        cog = cast("Select | None", self.get_cog("Select"))
        if cog is None:
            return log.warning(
                "Cannot set message callback for %d, Select cog not loaded",
                user_id,
            )

        return cog.set_message_callback(guild_id, user_id, callback)

    async def _maybe_load_jishaku(self) -> None:
        if not self.config.bot.allow_jishaku:
            return

        try:
            version = importlib.metadata.version("jishaku")
        except importlib.metadata.PackageNotFoundError:
            pass
        else:
            await self.load_extension("jishaku")
            log.info("Loaded jishaku extension (version: %s)", version)

    def refresh_config(self) -> Settings:
        config = self._config_refresher()
        self.config = config
        return config

    async def setup_hook(self) -> None:
        self.config.db.path.parent.mkdir(parents=True, exist_ok=True)
        with sqlite3.connect(self.config.db.path) as conn:
            self._run_config_pragmas(conn)
            run_default_migrations(conn)

        for path in self.config.bot.extensions:
            await self.load_extension(path, package=__package__)
        log.info("Loaded %d extensions", len(self.config.bot.extensions))
        await self._maybe_load_jishaku()

        await self.tree.set_translator(GettextTranslator())

        if self.startup_flags & StartupFlags.SYNC:
            commands = await self.tree.sync()
            log.info("Synced %d application commands", len(commands))

        invite_link = self.get_standard_invite()
        log.info("Invite link:\n%s", invite_link)

        if self.startup_flags & StartupFlags.CLOSE:
            sys.exit()

    def get_standard_invite(self) -> str:
        assert self.application is not None
        return discord.utils.oauth_url(
            self.application.id,
            scopes=("bot",),
            permissions=discord.Permissions(
                read_messages=True,
                send_messages=True,
                create_private_threads=True,
                send_messages_in_threads=True,
                embed_links=True,
                attach_files=True,
            ),
        )


class Context(commands.Context[Bot]): ...



import string
import platformdirs

APP_DIRS = platformdirs.PlatformDirs(
    appname="theticketbot",
    appauthor="thegamecracks",
    opinion=True,
)


def expand_app_dirs(path: str, *, strict: bool = False) -> str:
    template = string.Template(path)
    paths = {
        "USER_CONFIG_DIR": APP_DIRS.user_config_dir,
        "USER_DATA_DIR": APP_DIRS.user_data_dir,
        "USER_LOG_DIR": APP_DIRS.user_log_dir,
        "USER_RUNTIME_DIR": APP_DIRS.user_runtime_dir,
        "USER_STATE_DIR": APP_DIRS.user_state_dir,
    }

    if strict:
        return template.substitute(paths)

    return template.safe_substitute(paths)



import datetime
import json
import logging
import logging.handlers
from typing import Any

import discord

from .appdirs import APP_DIRS

LOG_RECORD_ATTRIBUTES = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",  # cached by formatException()
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "message",
    "module",
    "msecs",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}

log = logging.getLogger(__package__)


class JSONFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        data = self._prepare_record(record)
        return json.dumps(data, default=str)

    def _prepare_record(self, record: logging.LogRecord) -> dict[str, Any]:
        data = {}

        for k, v in vars(record).items():
            if k not in LOG_RECORD_ATTRIBUTES:
                data[k] = v

        created = datetime.datetime.fromtimestamp(
            record.created,
            tz=datetime.timezone.utc,
        )

        data["created"] = created.isoformat()
        data["level"] = record.levelname
        data["name"] = record.name
        data["message"] = record.getMessage()

        if record.exc_info is not None:
            data["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            data["stack_info"] = self.formatStack(record.stack_info)

        return data


def configure_logging(verbose: int) -> None:
    root_level = logging.INFO
    if verbose > 0:
        log.setLevel(logging.DEBUG)
    if verbose > 1:
        root_level = logging.DEBUG

    discord.utils.setup_logging(level=root_level, root=True)

    APP_DIRS.user_log_path.mkdir(parents=True, exist_ok=True)
    handler = logging.handlers.RotatingFileHandler(
        str(APP_DIRS.user_log_path / f"{__package__}.jsonl"),
        maxBytes=5_000_000,
        backupCount=3,
        encoding="utf-8",
    )
    handler.setFormatter(JSONFormatter())
    logging.getLogger().addHandler(handler)



import importlib.resources
import logging
import re
import sqlite3
from contextlib import contextmanager
from typing import Iterable, Iterator, NamedTuple, Self


log = logging.getLogger(__name__)


class Migration(NamedTuple):
    version: int
    sql: str


class Migrations(tuple[Migration, ...]):
    def after_version(self, version: int) -> Self:
        """Return a copy of self with only migrations after the given version."""
        return type(self)(m for m in self if m.version > version)

    def version_exists(self, version: int) -> bool:
        return any(m.version == version for m in self)

    @classmethod
    def from_iterable_unsorted(cls, it: Iterable[Migration]) -> Self:
        return cls(sorted(it, key=lambda m: m.version))


class MigrationFinder:
    _FILE_PATTERN = re.compile(r"(\d+)-(.+)\.sql")

    def discover(self) -> Migrations:
        migrations: list[Migration] = [Migration(version=-1, sql="")]

        assert __package__ is not None
        path = importlib.resources.files(__package__).joinpath("migrations/")

        for file in path.iterdir():
            if not file.is_file():
                continue

            m = self._FILE_PATTERN.fullmatch(file.name)
            if m is None:
                continue

            version = int(m[1])
            sql = file.read_text("utf-8")
            migrations.append(Migration(version=version, sql=sql))

        return Migrations.from_iterable_unsorted(migrations)


class Migrator:
    def __init__(self, conn: sqlite3.Connection) -> None:
        self.conn = conn

    def run_migrations(self, migrations: Migrations) -> None:
        version = self.get_version()
        if version > 0 and not migrations.version_exists(version):
            log.warning(
                "Unrecognized database version %d, skipping migrations",
                version,
            )
            return

        with self.begin() as conn:
            for version, script in migrations.after_version(version):
                log.info("Migrating database to v%d", version)
                conn.executescript(script)

            conn.execute(f"PRAGMA user_version = {version:d}")

    def get_version(self) -> int:
        version = self.conn.execute("PRAGMA user_version").fetchone()[0]
        log.debug("PRAGMA user_version returned %d", version)
        return version

    @contextmanager
    def begin(self) -> Iterator[sqlite3.Connection]:
        self.conn.execute("BEGIN")
        try:
            yield self.conn
        except BaseException:
            self.conn.rollback()
            raise
        else:
            self.conn.commit()


def run_default_migrations(conn: sqlite3.Connection) -> None:
    migrations = MigrationFinder().discover()
    migrator = Migrator(conn)
    migrator.run_migrations(migrations)



from __future__ import annotations

import functools
import logging
import re
import string
import time
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Iterable,
    Sequence,
    TypedDict,
    cast,
)

import asqlite
import discord
import discord.http
import humanize
from discord import app_commands
from discord.app_commands import locale_str as _
from discord.ext import commands, tasks

from theticketbot.bot import Bot
from theticketbot.database import DatabaseClient
from theticketbot.errors import AppCommandResponse
from theticketbot.translator import translate

if TYPE_CHECKING:
    from .select import MessageCallback

DEFAULT_STARTER_CONTENT = "$author $staff"
DEFAULT_TICKET_NAME = "$year-$month-$day $author"
MENTION_PATTERN = re.compile(r"<(@|@&)(\d+)>")

InboxRatelimit = Callable[[discord.Message, discord.Member], Awaitable[float]]

log = logging.getLogger(__name__)


class InboxView(discord.ui.View):
    def __init__(self, bot: Bot, ratelimit_check: InboxRatelimit) -> None:
        super().__init__(timeout=None)
        self.bot = bot
        self.ratelimit_check = ratelimit_check

    async def localize(self, locale: discord.Locale) -> None:
        async def t(s: _) -> str:
            return await translate(s, self.bot, locale=locale)

        # Button label for creating a new ticket
        self.create_ticket.label = await t(_("Create Ticket"))

    @discord.ui.button(custom_id="create-ticket", style=discord.ButtonStyle.primary)
    async def create_ticket(
        self,
        interaction: discord.Interaction,
        button: discord.ui.Button,
    ):
        # FIXME: this function is too big, can we do any better?
        assert isinstance(interaction.channel, discord.TextChannel)
        assert isinstance(interaction.user, discord.Member)
        assert interaction.guild is not None
        assert interaction.message is not None

        guild = interaction.guild
        message = interaction.message

        async with self.bot.acquire() as conn:
            # If the database was wiped, this will fail.
            row = await conn.fetchone("SELECT 1 FROM inbox WHERE id = ?", message.id)
            if row is None:
                content = _(
                    # Message sent when an inbox is not recognized
                    "Sorry, this inbox is no longer recognized and must be "
                    "re-created. Please notify a server admin!"
                )
                content = await translate(content, interaction)
                return await interaction.response.send_message(content, ephemeral=True)

            tickets = await self.get_active_user_tickets(
                interaction.channel.threads,
                conn,
                message.id,
                interaction.user.id,
            )
            tickets.sort(key=lambda t: t.id)
            max_tickets = await self.get_max_tickets(conn, message.id)

        if max_tickets > 0 and len(tickets) >= max_tickets:
            content = _(
                # Message sent when trying to create too many tickets
                # {0}: the ticket's link
                "You have too many tickets in this inbox. "
                "Please close your last ticket {0} before creating a new one."
            )
            content = await translate(content, interaction)
            content = content.format(tickets[-1].jump_url)
            return await interaction.response.send_message(content, ephemeral=True)

        retry_after = await self.ratelimit_check(message, interaction.user)
        if retry_after > 0:
            # Message sent when user is being ratelimited for an inbox
            # {0}: the duration in seconds to wait before retrying
            content = _("You are creating tickets too quickly! Please wait {0:.0f}s.")
            content = await translate(content, interaction)
            content = content.format(retry_after)
            return await interaction.response.send_message(content, ephemeral=True)

        # Message sent when creating a ticket
        content = await translate(_("Creating ticket..."), interaction)
        await interaction.response.send_message(content, ephemeral=True)

        async with self.bot.acquire() as conn:
            query = DatabaseClient(conn)
            ticket_name = await query.get_inbox_default_ticket_name(message.id)
            ticket_name = ticket_name or DEFAULT_TICKET_NAME

            # NOTE: counter may skip if thread creation fails
            counter = await query.increment_inbox_counter(message.id)

        created_at = interaction.created_at
        ticket_name = string.Template(ticket_name).safe_substitute(
            year=created_at.year,
            month=str(created_at.month).zfill(2),
            day=str(created_at.day).zfill(2),
            author=interaction.user.display_name,
            counter=str(counter % 10**4).zfill(4),
        )

        # Audit log reason for a user creating a ticket
        # {0}: the user's name
        reason = _("Ticket created by {0}")
        reason = await translate(reason, interaction, locale=guild.preferred_locale)
        reason = reason.format(interaction.user.name)

        try:
            ticket = await interaction.channel.create_thread(
                name=ticket_name[:100],
                invitable=False,
                reason=reason,
            )

            async with self.bot.acquire() as conn:
                query = DatabaseClient(conn)
                await query.add_ticket(
                    ticket_id=ticket.id,
                    inbox_id=message.id,
                    owner_id=interaction.user.id,
                    guild_id=guild.id,
                )

            async with self.bot.acquire() as conn:
                query = DatabaseClient(conn)
                mentions = await get_and_filter_inbox_staff(query, guild, message.id)
                mentions = " ".join(mentions)

                starter_content = await query.get_inbox_starter_content(message.id)
                starter_content = starter_content or DEFAULT_STARTER_CONTENT

            content = string.Template(starter_content).safe_substitute(
                author=interaction.user.mention,
                staff=mentions,
            )
            await ticket.send(content[:2000])
        except discord.Forbidden:
            content = _(
                # Message sent when creating a ticket failed due to insufficient permissions
                "I am missing the permissions needed to create a ticket here. "
                "Please notify a server admin!"
            )
            content = await translate(content, interaction)
            return await interaction.edit_original_response(content=content)
        except Exception:
            # Message sent when creating a ticket failed unexpectedly
            content = _("An unexpected error occurred while creating the ticket.")
            content = await translate(content, interaction)
            await interaction.edit_original_response(content=content)
            raise
        else:
            # Message sent after successfully creating a ticket
            # {0}: the ticket's link
            content = _("Your ticket is ready! {0}")
            content = await translate(content, interaction)
            content = content.format(ticket.jump_url)
            await interaction.edit_original_response(content=content)

    async def get_active_user_tickets(
        self,
        threads: Iterable[discord.Thread],
        conn: asqlite.Connection,
        inbox_id: int,
        owner_id: int,
    ) -> list[discord.Thread]:
        # Should do an intersect here, but this should be small enough
        c = await conn.execute(
            "SELECT id FROM ticket WHERE inbox_id = ? AND owner_id = ?",
            inbox_id,
            owner_id,
        )
        ticket_ids: set[int] = {row[0] for row in await c.fetchall()}

        members_intent = self.bot.intents.members
        active: list[discord.Thread] = []
        for t in threads:
            if t.id not in ticket_ids:
                continue
            if members_intent and discord.utils.get(t.members, id=owner_id) is None:
                continue
            active.append(t)
        return active

    async def get_max_tickets(self, conn: asqlite.Connection, inbox_id: int) -> int:
        row = await conn.fetchone(
            "SELECT max_tickets_per_user FROM inbox WHERE id = ?",
            inbox_id,
        )
        assert row is not None
        return row[0]


async def get_and_filter_inbox_staff(
    query: DatabaseClient,
    guild: discord.Guild,
    inbox_id: int,
) -> list[str]:
    mentions = await query.get_inbox_staff(inbox_id)
    return await filter_and_update_inbox_staff(query, guild, inbox_id, mentions)


async def filter_and_update_inbox_staff(
    query: DatabaseClient,
    guild: discord.Guild,
    inbox_id: int,
    mentions: Sequence[str],
) -> list[str]:
    role_mentions = set(m for m in mentions if m.startswith("<@&"))
    current_roles = {role.mention for role in guild.roles}
    removed = role_mentions - current_roles

    for mention in removed:
        await query.remove_inbox_staff(inbox_id, mention)

    return [m for m in mentions if m not in removed]


class InboxStaffView(discord.ui.View):
    def __init__(self, bot: Bot, inbox_id: int, staff: set[str]) -> None:
        super().__init__(timeout=None)
        self.bot = bot
        self.inbox_id = inbox_id
        self.staff = staff
        self.refresh()

    def refresh(self) -> None:
        staff = [mention_to_snowflake(staff) for staff in self.staff]
        self.on_staff_select.default_values = staff

    @discord.ui.select(cls=discord.ui.MentionableSelect, min_values=0, max_values=10)
    async def on_staff_select(
        self,
        interaction: discord.Interaction,
        select: discord.ui.MentionableSelect,
    ):
        ordered_mentions = [m.mention for m in select.values]
        mentions = set(ordered_mentions)
        added = mentions - self.staff
        removed = self.staff - mentions

        if len(added) == 0 and len(removed) == 0:
            # Message sent when submitting no changes to inbox staff
            content = _("You have not made any changes!")
            content = await translate(content, interaction)
            return await interaction.response.send_message(content, ephemeral=True)

        async with self.bot.acquire() as conn:
            query = DatabaseClient(conn)
            for mention in added:
                await query.add_inbox_staff(self.inbox_id, mention)
            for mention in removed:
                await query.remove_inbox_staff(self.inbox_id, mention)

        self.staff = mentions
        await interaction.response.defer()


def mention_to_snowflake(mention: str) -> discord.Object:
    m = MENTION_PATTERN.fullmatch(mention)
    if m is None:
        raise ValueError(f"Invalid user/role mention: {mention!r}")

    if m[1] == "@":
        return discord.Object(int(m[2]), type=discord.User)
    elif m[1] == "@&":
        return discord.Object(int(m[2]), type=discord.Role)
    raise RuntimeError(f"Unsupported mention type {m[1]!r}")


def snowflake_to_mention(obj: discord.Member | discord.Role | discord.Object) -> str:
    if not isinstance(obj, discord.Object):
        return obj.mention
    elif obj.type is discord.abc.User:
        return f"<@{obj.id}>"
    elif issubclass(obj.type, discord.Role):
        return f"<@&{obj.id}>"
    raise ValueError(f"Unsupported object type {obj.type}")


class SetInboxStarterContentModal(discord.ui.Modal, title="Starter Message"):
    content = discord.ui.TextInput(
        label="Content",
        style=discord.TextStyle.long,
        max_length=2000,
        required=False,
    )

    def __init__(self, bot: Bot, inbox: discord.Message) -> None:
        super().__init__()
        self.bot = bot
        self.inbox = inbox

    async def localize(self, locale: discord.Locale) -> None:
        async def t(s: _) -> str:
            return await translate(s, self.bot, locale=locale)

        # Modal title for changing an inbox's starter message
        self.title = await t(_("Starter Message"))
        # Modal text input label for an inbox's starter message content
        self.content.label = await t(_("Content"))

    async def set_defaults(self, conn: asqlite.Connection) -> None:
        query = DatabaseClient(conn)
        starter_content = await query.get_inbox_starter_content(self.inbox.id)
        starter_content = starter_content or DEFAULT_STARTER_CONTENT
        self.content.default = starter_content

    async def on_submit(self, interaction: discord.Interaction):
        async with self.bot.acquire() as conn:
            query = DatabaseClient(conn)
            await query.set_inbox_starter_content(self.inbox.id, self.content.value)

        # Message sent when an inbox's starter message is successfully changed
        # {0}: the inbox's link
        content = _("{0} 's starting message has been set!")
        content = await translate(content, interaction)
        content = content.format(self.inbox.jump_url)
        await interaction.response.send_message(content, ephemeral=True)


class SetTicketDefaultsModal(discord.ui.Modal, title="New Tickets"):
    name = discord.ui.TextInput(label="Name", max_length=100, required=False)

    def __init__(self, bot: Bot, inbox: discord.Message) -> None:
        super().__init__()
        self.bot = bot
        self.inbox = inbox

    async def localize(self, locale: discord.Locale) -> None:
        async def t(s: _) -> str:
            return await translate(s, self.bot, locale=locale)

        # Modal title for changing an inbox's defaults for new tickets
        self.title = await t(_("New Tickets"))
        # Modal text input label for ticket names
        self.name.label = await t(_("Name"))

    async def set_defaults(self, conn: asqlite.Connection) -> None:
        query = DatabaseClient(conn)
        ticket_name = await query.get_inbox_default_ticket_name(self.inbox.id)
        ticket_name = ticket_name or DEFAULT_TICKET_NAME
        self.name.default = ticket_name

    async def on_submit(self, interaction: discord.Interaction):
        async with self.bot.acquire() as conn:
            query = DatabaseClient(conn)
            await query.set_inbox_default_ticket_name(self.inbox.id, self.name.value)

        # Message sent when an inbox's ticket defaults were successfully changed
        # {0}: the inbox's link
        content = _("{0} 's ticket defaults have been set!")
        content = await translate(content, interaction)
        content = content.format(self.inbox.jump_url)
        await interaction.response.send_message(content, ephemeral=True)


def looks_like_an_inbox(bot: Bot, message: discord.Message) -> bool:
    # Good enough as a heuristic
    assert bot.user is not None
    return message.author.id == bot.user.id and len(message.components) > 0


class InboxMessageParams(TypedDict):
    embeds: list[discord.Embed]
    files: list[discord.File]


@app_commands.default_permissions(manage_guild=True)
@app_commands.guild_only()
class Inbox(
    commands.GroupCog,
    # Command group name
    # (alternatively translated as "panel")
    group_name=_("inbox"),
    # Command group description ("inbox")
    group_description=_("Manage the server's ticket inboxes."),
):
    _inbox_ratelimits: dict[tuple[int, int], tuple[float, app_commands.Cooldown]]

    def __init__(self, bot: Bot):
        self.bot = bot

        self._global_inbox_view = self.create_inbox_view()
        self._inbox_views: dict[int, InboxView] = {}
        self._inbox_ratelimits = {}

        self.cleanup_loop.start()
        self.bot.add_view(self._global_inbox_view)

    def create_inbox_view(self) -> InboxView:
        return InboxView(self.bot, ratelimit_check=self.check_ratelimit)

    async def cog_unload(self) -> None:
        self._global_inbox_view.stop()
        for view in self._inbox_views.values():
            view.stop()

    async def check_ratelimit(
        self,
        inbox: discord.Message,
        member: discord.Member,
    ) -> float:
        assert isinstance(inbox.channel, discord.TextChannel)

        key = (inbox.id, member.id)
        per = max(60.0, inbox.channel.slowmode_delay)
        per_cooldown = self._inbox_ratelimits.get(key)

        # Allow cooldowns to reset when slowmode is changed
        if per_cooldown is None or per != per_cooldown[0]:
            per_cooldown = (per, app_commands.Cooldown(1, per))
            self._inbox_ratelimits[key] = per_cooldown

        per, cooldown = per_cooldown
        return cooldown.update_rate_limit(time.monotonic()) or 0

    def set_inbox_callback(
        self,
        interaction: discord.Interaction,
        callback: MessageCallback,
    ) -> None:
        async def wrapper(interaction: discord.Interaction, message: discord.Message):
            # As long as message callbacks are guild-specific,
            # this assert should always be true
            assert interaction.guild == original_interaction.guild

            await self.check_inbox_message(interaction, message)
            await callback(interaction, message)

        assert interaction.guild is not None
        original_interaction = interaction
        self.bot.set_message_callback(
            interaction.guild.id,
            interaction.user.id,
            wrapper,
        )

    async def check_inbox_message(
        self,
        interaction: discord.Interaction,
        message: discord.Message,
    ) -> None:
        assert isinstance(interaction.user, discord.Member)

        async with self.bot.acquire() as conn:
            row = await conn.fetchone("SELECT 1 FROM inbox WHERE id = ?", message.id)

        if row is None:
            if looks_like_an_inbox(self.bot, message):
                content = _(
                    # Message sent when selecting a non-inbox message
                    # that looks like the message used to be an inbox
                    # {0}: the message's link
                    "Sorry, {0} is no longer recognized as an inbox "
                    "and must be re-created."
                )
            else:
                content = _(
                    # Message sent when selecting a non-inbox message
                    # {0}: the message's link
                    "Sorry, {0} is not an inbox. The message you select should have "
                    "a **Create Ticket** button under it."
                )

            content = await translate(content, interaction)
            content = content.format(message.jump_url)
            raise AppCommandResponse(content)

    @app_commands.command(
        # Subcommand name ("inbox")
        name=_("create"),
        # Subcommand description ("inbox create")
        description=_("Create a new inbox."),
    )
    @app_commands.rename(
        # Subcommand parameter name ("inbox create")
        channel=_("channel"),
    )
    @app_commands.describe(
        # Subcommand parameter description ("inbox create <channel>")
        channel=_("The channel to post the inbox."),
    )
    async def create(
        self,
        interaction: discord.Interaction,
        channel: discord.TextChannel,
    ):
        assert interaction.guild is not None

        # TODO: limit number of inboxes per guild

        bot_permissions = channel.permissions_for(interaction.guild.me)
        required = discord.Permissions(
            view_channel=True,
            send_messages=True,
            create_private_threads=True,
            send_messages_in_threads=True,
        )
        missing = bot_permissions & required ^ required
        if missing:
            missing = ", ".join(f"`{name}`" for name, value in missing if value)
            # Message sent when attempting to create an inbox with insufficient permissions
            # {0}: the channel's mention
            # {1}: a list of permissions that are missing
            content = _("I need the following permissions in {0}: {1}")
            content = await translate(content, interaction)
            content = content.format(channel.mention, missing)
            return await interaction.response.send_message(content, ephemeral=True)

        content = _(
            # Message sent when the user is creating a new inbox in a channel,
            # and the inbox needs a message to be included
            # {0}: the channel's mention
            "The {0} channel has been set as the destination for your new inbox. "
            "You must now select the message you want your inbox to have. "
            "To do this, right click or long tap a message, then open Apps "
            "and pick the *Select this message* command."
        )
        content = await translate(content, interaction)
        content = content.format(channel.mention)
        await interaction.response.send_message(content, ephemeral=True)
        self.bot.set_message_callback(
            interaction.guild.id,
            interaction.user.id,
            functools.partial(self.create_inbox, channel=channel),
        )

    async def create_inbox(
        self,
        interaction: discord.Interaction,
        message: discord.Message,
        channel: discord.TextChannel,
    ):
        assert interaction.guild is not None

        kwargs = await self.create_inbox_message(
            interaction,
            message,
            embed_url=channel.jump_url,
        )
        if kwargs is None:
            return

        view = self.create_inbox_view()
        await view.localize(interaction.guild.preferred_locale)

        message = await channel.send(view=view, **kwargs)
        assert message.guild is not None

        self._inbox_views[message.id] = view

        async with self.bot.acquire() as conn:
            query = DatabaseClient(conn)
            await query.add_inbox(
                message.id,
                message.channel.id,
                guild_id=message.guild.id,
            )

            starter_content = await translate(
                # The default starter message content for new tickets
                _("$author Thank you for creating a ticket!\n$staff"),
                interaction,
                locale=interaction.guild.preferred_locale,
            )
            await query.set_inbox_starter_content(message.id, starter_content)

            for staff in self.get_default_inbox_staff(channel):
                mention = snowflake_to_mention(staff)
                await query.add_inbox_staff(message.id, mention)

        # Message sent after a user creates an inbox
        # {0}: the inbox's link
        content = await translate(_("Your inbox has been created! {0}"), interaction)
        content = content.format(message.jump_url)
        await interaction.followup.send(content, ephemeral=True)

    async def create_inbox_message(
        self,
        interaction: discord.Interaction,
        message: discord.Message,
        *,
        embed_url: str | None = None,
    ) -> InboxMessageParams | None:
        embeds = [discord.Embed()]
        files: list[discord.File] = []

        if message.content != "":
            embeds[0].description = message.content

        max_attachment_size = self.bot.config.bot.inbox.max_attachment_size
        if sum(a.size for a in message.attachments) > max_attachment_size:
            content = _(
                # Message sent when attempting to create an inbox with too large attachments
                # {0}: the maximum cumulative filesize
                "The message's attachments are too large! "
                "The total size must be under {0}."
            )
            content = await translate(content, interaction)
            content = content.format(humanize.naturalsize(max_attachment_size))
            await interaction.response.send_message(content, ephemeral=True)
            return

        embeds_copied = False
        if len(embeds[0]) == 0 and len(message.embeds) > 0:
            embeds.clear()
            embeds.extend(embed.copy() for embed in message.embeds)
            embeds_copied = True

        await interaction.response.defer(ephemeral=True)

        url = embed_url or message.channel.jump_url
        for attachment in message.attachments:
            f = await attachment.to_file()
            files.append(f)
            image_url = f"attachment://{f.filename}"

            if embeds_copied:
                pass
            elif embeds[0].url is None:
                embeds[0].url = url
                embeds[0].set_image(url=image_url)
            else:
                embeds.append(discord.Embed(url=url).set_image(url=image_url))

        return {"embeds": embeds, "files": files}

    def get_default_inbox_staff(
        self,
        channel: discord.TextChannel,
    ) -> list[discord.Member | discord.Role | discord.Object]:
        assert self.bot.user is not None
        return [
            target
            for target, overwrite in channel.overwrites.items()
            if overwrite.manage_threads and target.id != self.bot.user.id
        ]

    @app_commands.command(
        # Subcommand name ("inbox")
        name=_("message"),
        # Subcommand description ("inbox message")
        description=_("Edit the message for an inbox."),
    )
    async def message(self, interaction: discord.Interaction):
        content = _(
            # Message sent when a user is editing an inbox's message,
            # and an inbox needs to be selected
            "You must now select the inbox you want to edit. "
            "To do this, right click or long tap a message, then open Apps "
            "and pick the *Select this message* command."
        )
        content = await translate(content, interaction)
        await interaction.response.send_message(content, ephemeral=True)
        self.set_inbox_callback(interaction, self.select_message_to_edit_inbox)

    async def select_message_to_edit_inbox(
        self,
        interaction: discord.Interaction,
        inbox: discord.Message,
    ):
        assert interaction.guild is not None
        guild_id = interaction.guild.id
        user_id = interaction.user.id

        # Message sent when a user is editing an inbox's message,
        # and a second message needs to be selected to copy its contents
        # {0}: the inbox's link
        content = _("{0} will be edited. Please select the message you want to copy.")
        content = await translate(content, interaction)
        content = content.format(inbox.jump_url)
        await interaction.response.send_message(content, ephemeral=True)
        callback = functools.partial(self.edit_inbox_message, inbox=inbox)
        self.bot.set_message_callback(guild_id, user_id, callback)

    async def edit_inbox_message(
        self,
        interaction: discord.Interaction,
        message: discord.Message,
        inbox: discord.Message,
    ):
        assert interaction.guild is not None

        if message == inbox:
            content = _(
                # Message sent when a user tries to edit an inbox message with itself
                "The inbox message cannot be edited with itself. "
                "Please select another message you want to copy."
            )
            raise AppCommandResponse(content)

        kwargs = await self.create_inbox_message(
            interaction,
            message,
            embed_url=inbox.channel.jump_url,
        )
        if kwargs is None:
            return

        kwargs = cast(dict[str, Any], kwargs)
        kwargs["attachments"] = kwargs.pop("files")

        # In case the guild locale changed, re-edit the view as well
        view = self.create_inbox_view()
        await view.localize(interaction.guild.preferred_locale)
        await inbox.edit(view=view, **kwargs)
        self._inbox_views[inbox.id] = view

        # Message sent after a user edits an inbox's message
        # {0}: the inbox's link
        content = await translate(_("{0} has been updated!"), interaction)
        content = content.format(inbox.jump_url)
        await interaction.followup.send(content, ephemeral=True)

    @app_commands.command(
        # Subcommand name ("inbox")
        name=_("staff"),
        # Subcommand description ("inbox staff")
        description=_("Manage staff for an inbox."),
    )
    async def staff(self, interaction: discord.Interaction):
        content = _(
            # Message sent when a user is managing staff for an inbox,
            # and an inbox needs to be selected
            "You must now select the inbox you want to manage staff for. "
            "To do this, right click or long tap a message, then open Apps "
            "and pick the *Select this message* command."
        )
        content = await translate(content, interaction)
        await interaction.response.send_message(content, ephemeral=True)
        self.set_inbox_callback(interaction, self.manage_inbox_staff)

    async def manage_inbox_staff(
        self,
        interaction: discord.Interaction,
        inbox: discord.Message,
    ):
        assert interaction.guild is not None
        async with self.bot.acquire() as conn:
            query = DatabaseClient(conn)
            staff = await get_and_filter_inbox_staff(query, interaction.guild, inbox.id)

        # Message sent above select menus when presenting an inbox's staff
        # {0}: the inbox's link
        content = _("Staff for {0} :")
        content = await translate(content, interaction)
        content = content.format(inbox.jump_url)
        view = InboxStaffView(self.bot, inbox.id, set(staff))

        return await interaction.response.send_message(
            content,
            ephemeral=True,
            view=view,
        )

    new_tickets = app_commands.Group(
        # Subcommand group name ("inbox")
        name=_("new-tickets"),
        # Subcommand group description ("inbox new-tickets")
        description=_("Manage new tickets created by an inbox."),
    )

    @new_tickets.command(
        # Subcommand name ("inbox new-tickets")
        name=_("starter"),
        # Subcommand description ("inbox new-tickets starter")
        description=_("Set the starting message for new tickets."),
    )
    async def new_tickets_starter(self, interaction: discord.Interaction):
        content = _(
            # Message sent when a user is changing the starter message for new tickets,
            # and an inbox needs to be selected
            "You must now select the inbox you want to edit. "
            "To do this, right click or long tap a message, then open Apps "
            "and pick the *Select this message* command."
        )
        content = await translate(content, interaction)
        await interaction.response.send_message(content, ephemeral=True)
        self.set_inbox_callback(interaction, self.edit_new_tickets_starter)

    async def edit_new_tickets_starter(
        self,
        interaction: discord.Interaction,
        inbox: discord.Message,
    ):
        modal = SetInboxStarterContentModal(self.bot, inbox)
        async with self.bot.acquire() as conn:
            await modal.set_defaults(conn)
        await modal.localize(interaction.locale)
        await interaction.response.send_modal(modal)

    @new_tickets.command(
        # Subcommand name ("inbox new-tickets")
        name=_("name"),
        # Subcommand description ("inbox new-tickets name")
        description=_("Set the name for new tickets."),
    )
    async def new_tickets_name(self, interaction: discord.Interaction):
        content = _(
            # Message sent when a user is changing the name for new tickets,
            # and an inbox needs to be selected
            "You must now select the inbox you want to edit. "
            "To do this, right click or long tap a message, then open Apps "
            "and pick the *Select this message* command."
        )
        content = await translate(content, interaction)
        await interaction.response.send_message(content, ephemeral=True)
        self.set_inbox_callback(interaction, self.edit_new_tickets_name)

    async def edit_new_tickets_name(
        self,
        interaction: discord.Interaction,
        inbox: discord.Message,
    ):
        modal = SetTicketDefaultsModal(self.bot, inbox)
        async with self.bot.acquire() as conn:
            await modal.set_defaults(conn)
        await modal.localize(interaction.locale)
        await interaction.response.send_modal(modal)

    @commands.Cog.listener("on_raw_thread_member_remove")
    async def on_ticket_owner_remove(self, payload: discord.RawThreadMembersUpdate):
        # NOTE: event may not fire on already-archived tickets
        guild = self.bot.get_guild(payload.guild_id)
        if guild is None:
            return log.warning("Ignoring unknown guild %d", payload.guild_id)

        thread = guild.get_thread(payload.thread_id)
        if thread is None:
            return log.warning("Ignoring unknown thread %d", payload.thread_id)

        if thread.owner_id != guild.me.id:
            return

        user_ids = set(map(int, payload.data.get("removed_member_ids", ())))

        async with self.bot.acquire() as conn:
            row = await conn.fetchone(
                "SELECT owner_id FROM ticket WHERE id = ?",
                payload.thread_id,
            )
            if row is None:
                return

        owner_id: int = row[0]
        if owner_id not in user_ids:
            return

        # Message sent when a user leaves their ticket
        # {0}: The ticket owner's mention
        content = _("Archiving ticket as the owner ({0}) has left the thread.")
        content = await translate(content, self.bot, locale=guild.preferred_locale)
        content = content.format(f"<@{owner_id}>")
        await self.archive_ticket_with_message(thread, content)

    @commands.Cog.listener("on_raw_member_remove")
    async def on_ticket_owner_remove_guild(self, payload: discord.RawMemberRemoveEvent):
        guild = self.bot.get_guild(payload.guild_id)
        if guild is None:
            return log.warning("Ignoring unknown guild %d", payload.guild_id)

        async with self.bot.acquire() as conn:
            rows = await conn.fetchall(
                "SELECT ticket.id FROM ticket "
                "JOIN inbox ON inbox.id = inbox_id "
                "JOIN message ON message.id = inbox.id "
                "JOIN channel ON channel.id = channel_id "
                "JOIN guild ON guild.id = guild_id "
                "WHERE guild_id = ? AND owner_id = ?",
                payload.guild_id,
                payload.user.id,
            )

        for (ticket_id,) in rows:
            thread = guild.get_thread(ticket_id)
            if thread is None:
                log.warning("Ignoring unknown thread %d", ticket_id)
                continue

            # Message sent when a user leaves a server with open tickets
            # {0}: The ticket owner's mention
            content = _("Archiving ticket as the owner ({0}) has left the server.")
            content = await translate(content, self.bot, locale=guild.preferred_locale)
            content = content.format(payload.user.mention)
            await self.archive_ticket_with_message(thread, content)

    async def archive_ticket_with_message(
        self,
        thread: discord.Thread,
        content: str,
    ) -> None:
        can_lock = thread.permissions_for(thread.guild.me).manage_threads
        await thread.send(content, allowed_mentions=discord.AllowedMentions.none())
        await thread.edit(archived=True, locked=can_lock)

    @commands.Cog.listener("on_raw_thread_update")
    async def lock_archived_tickets(self, payload: discord.RawThreadUpdateEvent):
        guild_id = payload.guild_id
        thread_id = payload.thread_id

        guild = self.bot.get_guild(guild_id)
        if guild is None:
            return log.warning("Ignoring unknown guild %d", guild_id)

        if int(payload.data["owner_id"]) != guild.me.id:
            return

        thread_metadata = payload.data["thread_metadata"]
        if not thread_metadata["archived"] or thread_metadata.get("locked"):
            return

        channel_id = int(payload.data["parent_id"])
        channel = guild.get_channel(channel_id)
        if channel is None:
            return log.warning(
                "Ignoring unknown parent %d for thread %d",
                channel_id,
                thread_id,
            )

        permissions = channel.permissions_for(guild.me)
        if not permissions.manage_threads:
            return

        async with self.bot.acquire() as conn:
            row = await conn.fetchone("SELECT 1 FROM ticket WHERE id = ?", thread_id)
            if row is None:
                return

        # Message sent when locking a thread after being archived
        content = _("This archived ticket will be locked to moderators only.")
        content = await translate(content, self.bot, locale=guild.preferred_locale)

        params = discord.http.handle_message_parameters(content)
        await self.bot.http.send_message(thread_id, params=params)
        await self.bot.http.edit_channel(thread_id, archived=True, locked=True)

    @tasks.loop(minutes=30)
    async def cleanup_loop(self) -> None:
        now = time.monotonic()

        to_remove: list[tuple[int, int]] = []
        for key, (_, cooldown) in self._inbox_ratelimits.items():
            if now > cooldown._last + cooldown.per:  # unfortunate reliance on internals
                to_remove.append(key)

        for key in to_remove:
            del self._inbox_ratelimits[key]


async def setup(bot: Bot):
    await bot.add_cog(Inbox(bot))



"""
This setup.py handles compiling .po (portable object) files into their
appropriate .mo (machine object) results.

Reference: https://setuptools.pypa.io/en/latest/userguide/extension.html#setuptools.command.build.SubCommand
Original: https://github.com/thegamecracks/discord.py-i18n-demo/blob/main/setup.py
"""

# pyright: strict
import os
import shutil
import subprocess
from pathlib import Path
from typing import Iterator
import warnings

from setuptools import Command, setup
from setuptools.command.build import SubCommand, build  # type: ignore  # missing stub


class build_mo(Command, SubCommand):
    """Builds machine object translation files."""

    build_lib = "build/lib"
    editable_mode = False  # shouldn't be necessary but pyright wants this

    _source_root = Path("src")

    # SubCommand protocol

    def initialize_options(self) -> None:
        """
        Set or (reset) all options/attributes/caches used by the command
        to their default values. Note that these values may be overwritten during
        the build.
        """

    def finalize_options(self) -> None:
        """
        Set final values for all options/attributes used by the command.
        Most of the time, each option/attribute/cache should only be set if it does not
        have any value yet (e.g. ``if self.attr is None: self.attr = val``).
        """

    def run(self) -> None:
        """
        Execute the actions intended by the command.
        (Side effects **SHOULD** only take place when ``run`` is executed,
        for example, creating new files or writing to the terminal output).
        """
        msgfmt = shutil.which("msgfmt")
        if msgfmt is None:
            if os.getenv("LOCALE_MANDATORY") == "1":
                raise RuntimeError(
                    "msgfmt not available for building localizations, "
                    "but is required due to LOCALE_MANDATORY=1"
                )

            return warnings.warn(
                "msgfmt is not available, building .mo files will be skipped",
                RuntimeWarning,
            )

        for source_po in self._find_po_files():
            output_po = self._get_output_path(source_po)
            output_mo = output_po.with_suffix(".mo")

            if not self.editable_mode:
                # Parent directory required for msgfmt to work correctly
                output_mo.parent.mkdir(parents=True, exist_ok=True)

            subprocess.check_call(["msgfmt", "-o", output_mo, output_po])

            if not self.editable_mode:
                # Space savings
                output_po.unlink()

    def get_source_files(self) -> list[str]:
        """
        Return a list of all files that are used by the command to create
        the expected outputs. For example, if your build command transpiles
        Java files into Python, you should list here all the Java files.
        The primary purpose of this function is to help populating the ``sdist``
        with all the files necessary to build the distribution.
        All files should be strings relative to the project root directory.
        """
        return list(self.get_output_mapping().values())

    def get_outputs(self) -> list[str]:
        """
        Return a list of files intended for distribution as they would have been
        produced by the build.
        These files should be strings in the form of
        ``"{build_lib}/destination/file/path"``.

        .. note::
           The return value of ``get_output()`` should include all files used as keys
           in ``get_output_mapping()`` plus files that are generated during the build
           and don't correspond to any source file already present in the project.
        """
        return list(self.get_output_mapping().keys())

    def get_output_mapping(self) -> dict[str, str]:
        """
        Return a mapping between destination files as they would be produced by the
        build (dict keys) into the respective existing (source) files (dict values).
        Existing (source) files should be represented as strings relative to the project
        root directory.
        Destination files should be strings in the form of
        ``"{build_lib}/destination/file/path"``.
        """
        mapping: dict[str, str] = {}
        for source_po in self._find_po_files():
            output_po = self._get_output_path(source_po)
            output_mo = output_po.with_suffix(".mo")
            mapping[str(output_mo)] = str(source_po)
        return mapping

    # Utility methods

    def _find_po_files(self) -> Iterator[Path]:
        """Yields all PO files in the project source directory."""
        for package_path in self._source_root.iterdir():
            if not package_path.is_dir():
                continue

            for po_path in package_path.rglob("*.po"):
                yield po_path

    def _get_output_path(self, path: Path) -> Path:
        if self.editable_mode:
            return path
        path = path.relative_to(self._source_root)
        return Path(self.build_lib) / path

    # Subcommand registration

    build.sub_commands.append(("build_mo", None))


setup(cmdclass={"build_mo": build_mo})



import discord
from discord import app_commands
from discord.ext import commands

from theticketbot.bot import Bot, Context


def count_localizations(command: app_commands.AppCommand) -> int:
    """Returns the number of localizations defined for a command.

    .. seealso::

        https://discord.com/developers/docs/interactions/application-commands#application-command-object

    """
    n = 0
    n += len(command.name_localizations)
    n += len(command.description_localizations)
    for option in command.options:
        n += len(option.name_localizations)
        n += len(option.description_localizations)
        if not isinstance(option, app_commands.AppCommandGroup):
            n += sum(len(choice.name_localizations) for choice in option.choices)
    return n


class Owner(commands.Cog):
    def __init__(self, bot: Bot):
        self.bot = bot

    async def cog_check(self, ctx: Context):  # type: ignore  # async is allowed
        return await commands.is_owner().predicate(ctx)

    @commands.command(name="reload-config", aliases=["config-reload"])
    async def reload_config(self, ctx: Context):
        """Reload the bot's configuration."""
        self.bot.refresh_config()
        await ctx.reply("Config reloaded!")

    @commands.command(name="sync")
    async def sync(self, ctx: Context, guild_id: int | None = None):
        """Synchronize the bot's application commands."""
        guild: discord.Object | None = None
        if guild_id is not None:
            guild = discord.Object(guild_id)

        commands = await ctx.bot.tree.sync(guild=guild)
        n_commands = len(commands)
        n_localizations = sum(map(count_localizations, commands))
        await ctx.send(
            f"{n_commands} command(s) synchronized "
            f"with a total of {n_localizations} localizations!"
        )


async def setup(bot: Bot):
    await bot.add_cog(Owner(bot))
